"""Configuration values for jax kernels."""

MEMORY_SAVING_CONFIGS = [
    'KERNEL_MATRIX_USE_LOOP',
    'NORMALIZED_KERNEL_PUSH_DOWN'
]

SAVE_MEMORY = False
"""bool: Apply multiple tricks to reduce memory overhead.

This could slow down computation significantly but may allow the direct
application to larger problems when running into GPU memory issues.

This variable activates all configs in `MEMORY_SAVING_CONFIGS`.
"""

KERNEL_MATRIX_USE_LOOP = True
"""bool: Use loop when constructing kernel matrix

Compute kernel matrix using a dynamically unrolled loop instead of
vectorization. Reduces memory overhead as not all intermediate tensors have to
be allocated at once.
"""

NORMALIZED_KERNEL_PUSH_DOWN = False
"""bool: Push down the computation of normalization factor

Moves the computation of normalization to each individual instance. This can
reduce memory overhead, especially when computing gradients with respect to
kernel parameters at the cost of higher computation time (approx. 3 fold).  In
this case k(x,x) and k(y,y) needed to be recomputed for each instance.
"""


def config_value(config_var_name):
    """Get value of a config variable.

    Also accounts for config variables that activate other variables such as
    `SAVE_MEMORY`.
    """
    if config_var_name in MEMORY_SAVING_CONFIGS and SAVE_MEMORY:
        return True
    return globals()[config_var_name]



"""Utilities for usage with strings."""
from collections import defaultdict

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


class AsciiBytesTransformer(TransformerMixin):
    """Convert python strings into ascii byte arrays.

    This allows them to be used with jax.
    """

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        if len(np.unique([len(x) for x in X])) == 1:
            # All strings are of same length
            return np.asarray([bytearray(x, 'ascii', 'strict') for x in X])

        # Varying length
        return [np.asarray(bytearray(a, 'ascii', 'strict')) for a in X]

    def inverse_transform(self, X):
        return [str(bytes(a), 'ascii', 'strict') for a in X]


class CompressAlphabetTransformer(TransformerMixin, BaseEstimator):
    """Determines which chars are used and maps them to values."""

    def __init__(self, output_dtype=np.uint8):
        self.output_dtype = output_dtype
        self._unique_chars = None
        self._mapping = None

    def fit(self, X, y=None):
        single_str = ''.join(X)
        self._unique_chars = list(set(single_str))
        self._unique_chars.sort()
        n_alphabet = len(self._unique_chars)
        self._mapping = defaultdict(
            lambda: n_alphabet,
            [(char, i) for i, char in enumerate(self._unique_chars)]
        )
        return self

    def transform(self, X, y=None):
        X_transf = [
            [self._mapping[char] for char in x]
            for x in X
        ]
        return np.asarray(X_transf, dtype=self.output_dtype)


def get_translation_table(input_symbols, output_symbols, mapping,
                          dtype=np.uint8):
    """Build a translation table for characters using mapping.

    Returns a array, where index i corresponding to input_symbol contains the
    matching value from output_symbols.

    Example:
        >>> get_translation_table(
                input_symbols=['a', 'b', 'c'],
                output_symbols=['c', 'b', 'a'],
                mapping = {'a': 0, 'b': 1, 'c': 2}
            )
        array([2, 1, 0])

    """
    input_transf = np.array([mapping[symb] for symb in input_symbols], dtype=dtype)
    assert len(np.unique(input_transf)) == len(input_symbols)
    output_transf = np.array([mapping[symb] for symb in output_symbols], dtype=dtype)
    input_order = np.argsort(input_transf)
    return output_transf[input_order]



class NGramTransformer(TransformerMixin, BaseEstimator):
    def __init__(self, ngram_length):
        self.ngram_length = ngram_length

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        n = X.shape[1]
        ngram_slices = [
            X[:, i:n+1-self.ngram_length+i]
            for i in range(0, self.ngram_length)
        ]
        return np.stack(ngram_slices, axis=2)



"""Subclass of sklearn Gaussian process classifier using JAX."""
from functools import partial
from sklearn.gaussian_process import GaussianProcessClassifier as GPC
from sklearn.gaussian_process._gpc import (
    _BinaryGaussianProcessClassifierLaplace)
from sklearn.multiclass import OneVsRestClassifier, OneVsOneClassifier
from sklearn.utils.validation import check_X_y

import numpy

import jax.numpy as np
from jax.scipy.linalg import cholesky, cho_solve, solve
from jax.scipy.special import expit
from jax import ops
from jax import jit


@partial(jit, static_argnums=0)
def _newton_iteration(y_train, K, f):
    pi = expit(f)
    W = pi * (1 - pi)
    # Line 5
    W_sr = np.sqrt(W)
    W_sr_K = W_sr[:, np.newaxis] * K
    B = np.eye(W.shape[0]) + W_sr_K * W_sr
    L = cholesky(B, lower=True)
    # Line 6
    b = W * f + (y_train - pi)
    # Line 7
    a = b - W_sr * cho_solve((L, True), W_sr_K.dot(b))
    # Line 8
    f = K.dot(a)

    # Line 10: Compute log marginal likelihood in loop and use as
    #          convergence criterion
    lml = -0.5 * a.T.dot(f) \
        - np.log1p(np.exp(-(y_train * 2 - 1) * f)).sum() \
        - np.log(np.diag(L)).sum()
    return lml, f, (pi, W_sr, L, b, a)


class BinaryGaussianProcessClassifier(_BinaryGaussianProcessClassifierLaplace):
    def _posterior_mode(self, K, return_temporaries=False):
        """Mode-finding for binary Laplace GPC and fixed kernel.
        This approximates the posterior of the latent function values for given
        inputs and target observations with a Gaussian approximation and uses
        Newton's iteration to find the mode of this approximation.
        """
        # Based on Algorithm 3.1 of GPML

        # If warm_start are enabled, we reuse the last solution for the
        # posterior mode as initialization; otherwise, we initialize with 0
        if self.warm_start and hasattr(self, "f_cached") \
           and self.f_cached.shape == self.y_train_.shape:
            f = self.f_cached
        else:
            f = np.zeros_like(self.y_train_, dtype=np.float32)

        # Use Newton's iteration method to find mode of Laplace approximation
        log_marginal_likelihood = -np.inf
        newton_iteration = partial(_newton_iteration, self.y_train_, K)

        for _ in range(self.max_iter_predict):
            lml, f, (pi, W_sr, L, b, a) = newton_iteration(f)
            # Check if we have converged (log marginal likelihood does
            # not decrease)
            # XXX: more complex convergence criterion
            if lml - log_marginal_likelihood < 1e-10:
                break
            log_marginal_likelihood = lml

        self.f_cached = f  # Remember solution for later warm-starts
        if return_temporaries:
            return log_marginal_likelihood, (pi, W_sr, L, b, a)
        else:
            return log_marginal_likelihood

    def log_marginal_likelihood(self, theta=None, eval_gradient=False,
                                clone_kernel=False):
        """Returns log-marginal likelihood of theta for training data.

        Parameters
        ----------
        theta : array-like of shape (n_kernel_params,) or None
            Kernel hyperparameters for which the log-marginal likelihood is
            evaluated. If None, the precomputed log_marginal_likelihood
            of ``self.kernel_.theta`` is returned.
        eval_gradient : bool, default: False
            If True, the gradient of the log-marginal likelihood with respect
            to the kernel hyperparameters at position theta is returned
            additionally. If True, theta must not be None.
        clone_kernel : bool, default=True
            If True, the kernel attribute is copied. If False, the kernel
            attribute is modified, but may result in a performance improvement.
        Returns
        -------
        log_likelihood : float
            Log-marginal likelihood of theta for training data.
        log_likelihood_gradient : array, shape = (n_kernel_params,), optional
            Gradient of the log-marginal likelihood with respect to the kernel
            hyperparameters at position theta.
            Only returned when eval_gradient is True.
        """

        if theta is None:
            if eval_gradient:
                raise ValueError(
                    "Gradient can only be evaluated for theta!=None")
            return self.log_marginal_likelihood_value_

        kernel_matrix_fn = self.kernel_.get_kernel_matrix_fn(eval_gradient)

        if eval_gradient:
            K, K_gradient = kernel_matrix_fn(theta, self.X_train_, None)
        else:
            K = kernel_matrix_fn(theta, self.X_train_, None)

        # Compute log-marginal-likelihood Z and also store some temporaries
        # which can be reused for computing Z's gradient
        Z, (pi, W_sr, L, b, a) = \
            self._posterior_mode(K, return_temporaries=True)

        if not eval_gradient:
            return Z

        # Compute gradient based on Algorithm 5.1 of GPML

        d_Z = np.empty(theta.shape[0])
        # XXX: Get rid of the np.diag() in the next line
        R = W_sr[:, np.newaxis] * cho_solve((L, True), np.diag(W_sr))  # Line 7
        C = solve(L, W_sr[:, np.newaxis] * K)  # Line 8
        # Line 9: (use einsum to compute np.diag(C.T.dot(C))))
        s_2 = -0.5 * (np.diag(K) - np.einsum('ij, ij -> j', C, C)) \
            * (pi * (1 - pi) * (1 - 2 * pi))  # third derivative

        for j in range(d_Z.shape[0]):
            C = K_gradient[:, :, j]   # Line 11
            # Line 12: (R.T.ravel().dot(C.ravel()) = np.trace(R.dot(C)))
            s_1 = .5 * a.T.dot(C).dot(a) - .5 * R.T.ravel().dot(C.ravel())

            b = C.dot(self.y_train_ - pi)  # Line 13
            s_3 = b - K.dot(R.dot(b))  # Line 14

            d_Z = ops.index_update(d_Z, j, s_1 + s_2.T.dot(s_3))  # Line 15

        return (
            numpy.asarray(Z, dtype=numpy.float64),
            numpy.asarray(d_Z, dtype=numpy.float64)
        )


class GaussianProcessClassifier(GPC):
    def fit(self, X, y):
        """Fit Gaussian process classification model
        Parameters
        ----------
        X : sequence of length n_samples
            Feature vectors or other representations of training data.
            Could either be array-like with shape = (n_samples, n_features)
            or a list of objects.
        y : array-like of shape (n_samples,)
            Target values, must be binary
        Returns
        -------
        self : returns an instance of self.
        """
        if self.kernel is None or self.kernel.requires_vector_input:
            X, y = check_X_y(X, y, multi_output=False,
                             ensure_2d=True, dtype="numeric")
        else:
            X, y = check_X_y(X, y, multi_output=False,
                             ensure_2d=False, dtype=None)

        self.base_estimator_ = BinaryGaussianProcessClassifier(
            self.kernel, self.optimizer, self.n_restarts_optimizer,
            self.max_iter_predict, self.warm_start, self.copy_X_train,
            self.random_state)

        self.classes_ = numpy.unique(y)
        self.n_classes_ = self.classes_.size
        if self.n_classes_ == 1:
            raise ValueError("GaussianProcessClassifier requires 2 or more "
                             "distinct classes; got %d class (only class %s "
                             "is present)"
                             % (self.n_classes_, self.classes_[0]))
        if self.n_classes_ > 2:
            if self.multi_class == "one_vs_rest":
                self.base_estimator_ = \
                    OneVsRestClassifier(self.base_estimator_,
                                        n_jobs=self.n_jobs)
            elif self.multi_class == "one_vs_one":
                self.base_estimator_ = \
                    OneVsOneClassifier(self.base_estimator_,
                                       n_jobs=self.n_jobs)
            else:
                raise ValueError("Unknown multi-class mode %s"
                                 % self.multi_class)

        self.base_estimator_.fit(X, y)

        if self.n_classes_ > 2:
            self.log_marginal_likelihood_value_ = numpy.mean(
                [estimator.log_marginal_likelihood()
                 for estimator in self.base_estimator_.estimators_])
        else:
            self.log_marginal_likelihood_value_ = \
                self.base_estimator_.log_marginal_likelihood()

        return self



"""Implementation of string kernels."""
from functools import partial
from jax import device_put, vmap
import jax.numpy as np
from jax.lax import dynamic_slice_in_dim
from jax.experimental import loops

from sklearn_jax_kernels import Kernel
from sklearn.gaussian_process.kernels import GenericKernelMixin, Hyperparameter


class SpectrumKernel(GenericKernelMixin, Kernel):
    """Spectrum string kernel.

    As described in:
    ```
    @incollection{leslie2001spectrum,
      title={
        The spectrum kernel: A string kernel for SVM protein classification},
      author={
        Leslie, Christina and Eskin, Eleazar and Noble, William Stafford},
      booktitle={Biocomputing 2002},
      pages={564--575},
      year={2001},
      publisher={World Scientific}
    }
    ```
    """

    def __init__(self, n_gram_length):
        """Spectrum kernel on strings.

        Assumes input was transformed via `AcsiiBytesTransformer` or similar
        tranformation into a jax compatible datatype.

        Parameters:
            n_gram_length: Length of ngrams to compare. If `None` it is assumed
                that the input is 2d where the final axis is the n_grams.

        """
        self.n_gram_length = n_gram_length

    @property
    def pure_kernel_fn(self):
        """Return the pure fuction for computing the kernel."""
        n_gram_length = self.n_gram_length

        def kmer_kernel_fn(theta, kmers1, kmers2):
            same_kmer = np.all(kmers1[None, :, :] == kmers2[:, None, :], axis=2)
            return np.sum(same_kmer)

        if n_gram_length is None:
            # Assume input is kmer transformed
            kernel_fn = kmer_kernel_fn
        else:
            def kernel_fn(theta, string1, string2):
                def make_to_kmers(string):
                    ngram_slices = [
                        string[i:len(string)+1-n_gram_length+i]
                        for i in range(0, n_gram_length)
                    ]
                    return np.stack(ngram_slices, axis=1)
                kmers1 = make_to_kmers(string1)
                kmers2 = make_to_kmers(string2)
                return kmer_kernel_fn(theta, kmers1, kmers2)
        return kernel_fn

    @property
    def hyperparameters(self):
        """Return a list of all hyperparameter."""
        return []

    @property
    def theta(self):
        """Return the (flattened, log-transformed) non-fixed hyperparameters.

        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.

        Returns:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        return np.empty((0,))

    @theta.setter
    def theta(self, theta):
        """Set the (flattened, log-transformed) non-fixed hyperparameters.

        Parameters:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """

    @property
    def bounds(self):
        """Return the log-transformed bounds on the theta.

        Returns:
            bounds : array, shape (n_dims, 2)
                The log-transformed bounds on the kernel's hyperparameters
                theta

        """
        return np.empty((0, 2))

    def is_stationary(self):
        """Whether this kernel is stationary."""
        return False


class DistanceSpectrumKernel(Kernel):
    """Spectrum kernel weighting ngrams by their distance in the sequence."""

    def __init__(self, distance_kernel: Kernel, n_gram_length):
        """Initialize DistanceSpectrumKernel using distance_kernel.

        Args:
            distance_kernel (Kernel): Kernel used to quantify distance of kmers
                in the string.

        """
        self.distance_kernel = distance_kernel
        self.n_gram_length = n_gram_length

    @property
    def pure_kernel_fn(self):
        """Return the pure function for computing the kernel."""
        n_gram_length = self.n_gram_length

        distance_kernel = self.distance_kernel.pure_kernel_fn

        def kmer_kernel_fn(theta, kmers1, kmers2):
            pos_kernel = partial(distance_kernel, theta)
            same_kmer = np.all(
                kmers1[:, None, :] == kmers2[None, :, :],
                axis=2
            )
            offsets1 = np.arange(kmers1.shape[0])
            offsets2 = np.arange(kmers2.shape[0])
            distance_weight = vmap(
                lambda i: vmap(lambda j: pos_kernel(i, j))(offsets2))(offsets1)
            return np.sum(same_kmer * distance_weight)

        if n_gram_length is None:
            # Assume input is kmer transformed
            kernel_fn = kmer_kernel_fn
        else:
            def kernel_fn(theta, string1, string2):
                def make_to_kmers(string):
                    ngram_slices = [
                        string[i:len(string)+1-n_gram_length+i]
                        for i in range(0, n_gram_length)
                    ]
                    return np.stack(ngram_slices, axis=1)
                kmers1 = make_to_kmers(string1)
                kmers2 = make_to_kmers(string2)
                return kmer_kernel_fn(theta, kmers1, kmers2)

        return kernel_fn

    @property
    def hyperparameters(self):
        """Return a list of all hyperparameter."""
        r = []
        for hyperparameter in self.distance_kernel.hyperparameters:
            r.append(Hyperparameter("distance_kernel__" + hyperparameter.name,
                                    hyperparameter.value_type,
                                    hyperparameter.bounds,
                                    hyperparameter.n_elements))
        return r

    @property
    def theta(self):
        """Return the (flattened, log-transformed) non-fixed hyperparameters.

        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.

        Returns:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        return self.distance_kernel.theta

    @theta.setter
    def theta(self, theta):
        """Set the (flattened, log-transformed) non-fixed hyperparameters.

        Parameters:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        self.distance_kernel.theta = theta

    @property
    def bounds(self):
        """Return the log-transformed bounds on the theta.

        Returns:
            bounds : array, shape (n_dims, 2)
                The log-transformed bounds on the kernel's hyperparameters
                theta

        """
        return self.distance_kernel.bounds

    def get_params(self, deep=True):
        """Get parameters of this kernel.

        Parameters:
            deep : boolean, optional
                If True, will return the parameters for this estimator and
                contained subobjects that are estimators.

        Returns:
            params : mapping of string to any
                Parameter names mapped to their values.

        """
        params = dict(
            distance_kernel=self.distance_kernel,
            n_gram_length=self.n_gram_length
        )
        if deep:
            deep_items = self.distance_kernel.get_params().items()
            params.update(
                ('distance_kernel__' + k, val) for k, val in deep_items)
        return params

    def __eq__(self, b):
        """Whether two instances are considered equal."""
        if type(self) != type(b):
            return False
        return (
            self.distance_kernel == b.distance_kernel and
            self.n_gram_length == b.n_gram_length
        )

    def is_stationary(self):
        """Whether this kernel is stationary."""
        return False


class DistanceFromEndSpectrumKernel(Kernel):
    """Spectrum kernel weighting ngrams by their distance in the sequence."""

    def __init__(self, distance_kernel: Kernel, n_gram_length):
        """Initialize DistanceSpectrumKernel using distance_kernel.

        Args:
            distance_kernel (Kernel): Kernel used to quantify distance of kmers
                in the string.

        """
        self.distance_kernel = distance_kernel
        self.n_gram_length = n_gram_length

    @property
    def pure_kernel_fn(self):
        """Return the pure fuction for computing the kernel."""
        n_gram_length = self.n_gram_length

        distance_kernel = self.distance_kernel.pure_kernel_fn

        def kmer_kernel_fn(theta, kmers1, kmers2):
            pos_kernel = partial(distance_kernel, theta)
            kmer2_offsets = np.arange(kmers1.shape[0], dtype=np.uint32)
            distances_from_end = np.min(
                np.stack([kmer2_offsets, kmer2_offsets[::-1]], axis=0), axis=0)

            with loops.Scope() as s:
                s.out = 0.
                for i in s.range(kmers1.shape[0]):
                    distance_from_end = distances_from_end[i]
                    kmer = kmers1[i]
                    distances = vmap(
                        lambda j: pos_kernel(distance_from_end, j))(distances_from_end)
                    is_same = np.all(kmer[None, :] == kmers2, axis=1)
                    n_matches = np.sum(is_same * distances)
                    s.out += n_matches
            return s.out
        if n_gram_length is None:
            # Assume input is kmer transformed
            kernel_fn = kmer_kernel_fn
        else:
            def kernel_fn(theta, string1, string2):
                def make_to_kmers(string):
                    ngram_slices = [
                        string[i:len(string)+1-n_gram_length+i]
                        for i in range(0, n_gram_length)
                    ]
                    return np.stack(ngram_slices, axis=1)
                kmers1 = make_to_kmers(string1)
                kmers2 = make_to_kmers(string2)
                return kmer_kernel_fn(theta, kmers1, kmers2)

        return kernel_fn

    @property
    def hyperparameters(self):
        """Return a list of all hyperparameter."""
        r = []
        for hyperparameter in self.distance_kernel.hyperparameters:
            r.append(Hyperparameter("distance_kernel__" + hyperparameter.name,
                                    hyperparameter.value_type,
                                    hyperparameter.bounds,
                                    hyperparameter.n_elements))
        return r

    @property
    def theta(self):
        """Return the (flattened, log-transformed) non-fixed hyperparameters.

        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.

        Returns:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        return self.distance_kernel.theta

    @theta.setter
    def theta(self, theta):
        """Set the (flattened, log-transformed) non-fixed hyperparameters.

        Parameters:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        self.distance_kernel.theta = theta

    @property
    def bounds(self):
        """Return the log-transformed bounds on the theta.

        Returns:
            bounds : array, shape (n_dims, 2)
                The log-transformed bounds on the kernel's hyperparameters
                theta

        """
        return self.distance_kernel.bounds

    def get_params(self, deep=True):
        """Get parameters of this kernel.

        Parameters:
            deep : boolean, optional
                If True, will return the parameters for this estimator and
                contained subobjects that are estimators.

        Returns:
            params : mapping of string to any
                Parameter names mapped to their values.

        """
        params = dict(
            distance_kernel=self.distance_kernel,
            n_gram_length=self.n_gram_length
        )
        if deep:
            deep_items = self.distance_kernel.get_params().items()
            params.update(
                ('distance_kernel__' + k, val) for k, val in deep_items)
        return params

    def __eq__(self, b):
        """Whether two instances are considered equal."""
        if type(self) != type(b):
            return False
        return (
            self.distance_kernel == b.distance_kernel and
            self.n_gram_length == b.n_gram_length
        )

    def is_stationary(self):
        """Whether this kernel is stationary."""
        return False


class RevComplementSpectrumKernel(GenericKernelMixin, Kernel):
    """Spectrum string kernel which also count reverse complement matches."""

    def __init__(self, n_gram_length, mapping):
        """Spectrum kernel on strings.

        Assumes input was transformed via `AcsiiBytesTransformer` or similar
        tranformation into a jax compatible datatype.

        Parameters:
            n_gram_length: Length of ngrams to compare. If `None` it is assumed
                that the input is 2d where the final axis is the n_grams.
            mapping: Array of length of alphabet which defines what is
                considered the complement to a particular character

        """
        self.n_gram_length = n_gram_length
        self.mapping = mapping

    @property
    def pure_kernel_fn(self):
        """Return the pure fuction for computing the kernel."""
        n_gram_length = self.n_gram_length
        mapping = device_put(np.array(self.mapping))

        def kmer_kernel_fn(theta, kmers1, kmers2):
            with loops.Scope() as s:
                s.out = 0.
                for i in s.range(kmers1.shape[0]):
                    kmer = kmers1[i]
                    rev_comp = mapping[kmer][::-1]
                    is_same_fw = np.all(kmer[None, :] == kmers2, axis=1)
                    is_same_rev_comp = np.all(
                        rev_comp[None, :] == kmers2, axis=1)
                    n_matches = np.sum(is_same_fw) + np.sum(is_same_rev_comp)
                    s.out += n_matches
            return s.out

        if n_gram_length is None:
            # Assume input is kmer transformed
            kernel_fn = kmer_kernel_fn
        else:
            def kernel_fn(theta, string1, string2):
                def make_to_kmers(string):
                    ngram_slices = [
                        string[i:len(string)+1-n_gram_length+i]
                        for i in range(0, n_gram_length)
                    ]
                    return np.stack(ngram_slices, axis=1)
                kmers1 = make_to_kmers(string1)
                kmers2 = make_to_kmers(string2)
                return kmer_kernel_fn(theta, kmers1, kmers2)
        return kernel_fn

    @property
    def hyperparameters(self):
        """Return a list of all hyperparameter."""
        return []

    @property
    def theta(self):
        """Return the (flattened, log-transformed) non-fixed hyperparameters.

        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.

        Returns:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        return np.empty((0,))

    @theta.setter
    def theta(self, theta):
        """Set the (flattened, log-transformed) non-fixed hyperparameters.

        Parameters:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """

    @property
    def bounds(self):
        """Return the log-transformed bounds on the theta.

        Returns:
            bounds : array, shape (n_dims, 2)
                The log-transformed bounds on the kernel's hyperparameters
                theta

        """
        return np.empty((0, 2))

    def is_stationary(self):
        """Whether this kernel is stationary."""
        return False


class DistanceRevComplementSpectrumKernel(Kernel):
    """Spectrum kernel weighting ngrams by their distance in the sequence."""

    def __init__(self, distance_kernel: Kernel, n_gram_length, mapping):
        """Initialize DistanceSpectrumKernel using distance_kernel.

        Args:
            distance_kernel (Kernel): Kernel used to quantify distance of kmers
                in the string.

        """
        self.distance_kernel = distance_kernel
        self.n_gram_length = n_gram_length
        self.mapping = mapping

    @property
    def pure_kernel_fn(self):
        """Return the pure function for computing the kernel."""
        n_gram_length = self.n_gram_length

        distance_kernel = self.distance_kernel.pure_kernel_fn

        def kmer_kernel_fn(theta, kmers1, kmers2):
            mapping = device_put(self.mapping)
            pos_kernel = partial(distance_kernel, theta)

            rev_complement = np.reshape(
                mapping[np.ravel(kmers2)], kmers2.shape)[:, ::-1]

            same_kmer = np.all(
                kmers1[:, None, :] == kmers2[None, :, :],
                axis=2
            )
            same_rev_comp = np.all(
                kmers1[:, None, :] == rev_complement[None, :, :],
                axis=2
            )
            offsets1 = np.arange(kmers1.shape[0])
            offsets2 = np.arange(kmers2.shape[0])
            weight = vmap(
                lambda i: vmap(lambda j: pos_kernel(i, j))(offsets2))(offsets1)
            weight_rev_comp = vmap(
                lambda i: vmap(lambda j: pos_kernel(i, j))(offsets2[::-1]))(offsets1)

            return (
                np.sum(same_kmer * weight) +
                np.sum(same_rev_comp * weight_rev_comp)
            )

        if n_gram_length is None:
            # Assume input is kmer transformed
            kernel_fn = kmer_kernel_fn
        else:
            def kernel_fn(theta, string1, string2):
                def make_to_kmers(string):
                    ngram_slices = [
                        string[i:len(string)+1-n_gram_length+i]
                        for i in range(0, n_gram_length)
                    ]
                    return np.stack(ngram_slices, axis=1)
                kmers1 = make_to_kmers(string1)
                kmers2 = make_to_kmers(string2)
                return kmer_kernel_fn(theta, kmers1, kmers2)

        return kernel_fn

    @property
    def hyperparameters(self):
        """Return a list of all hyperparameter."""
        r = []
        for hyperparameter in self.distance_kernel.hyperparameters:
            r.append(Hyperparameter("distance_kernel__" + hyperparameter.name,
                                    hyperparameter.value_type,
                                    hyperparameter.bounds,
                                    hyperparameter.n_elements))
        return r

    @property
    def theta(self):
        """Return the (flattened, log-transformed) non-fixed hyperparameters.

        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.

        Returns:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        return self.distance_kernel.theta

    @theta.setter
    def theta(self, theta):
        """Set the (flattened, log-transformed) non-fixed hyperparameters.

        Parameters:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        self.distance_kernel.theta = theta

    @property
    def bounds(self):
        """Return the log-transformed bounds on the theta.

        Returns:
            bounds : array, shape (n_dims, 2)
                The log-transformed bounds on the kernel's hyperparameters
                theta

        """
        return self.distance_kernel.bounds

    def get_params(self, deep=True):
        """Get parameters of this kernel.

        Parameters:
            deep : boolean, optional
                If True, will return the parameters for this estimator and
                contained subobjects that are estimators.

        Returns:
            params : mapping of string to any
                Parameter names mapped to their values.

        """
        params = dict(
            distance_kernel=self.distance_kernel,
            n_gram_length=self.n_gram_length,
            mapping=self.mapping
        )
        if deep:
            deep_items = self.distance_kernel.get_params().items()
            params.update(
                ('distance_kernel__' + k, val) for k, val in deep_items)
        return params

    def __eq__(self, b):
        """Whether two instances are considered equal."""
        if type(self) != type(b):
            return False
        return (
            self.distance_kernel == b.distance_kernel and
            self.n_gram_length == b.n_gram_length
        )

    def is_stationary(self):
        """Whether this kernel is stationary."""
        return False



"""Jax implementation of sklearn kernels."""
from .base_kernels import (
    ConstantKernel,
    Exponentiation,
    Kernel,
    KernelOperator,
    NormalizedKernel,
    Product,
    RBF,
    Sum
)
from .gpc import GaussianProcessClassifier

__all__ = [
    'ConstantKernel', 'Exponentiation', 'GaussianProcessClassifier', 'Kernel',
    'KernelOperator', 'NormalizedKernel', 'Product', 'RBF', 'Sum'
]
__version__ = '0.1.0'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sklearn_jax_kernels', 'sklearn_jax_kernels.structured']

package_data = \
{'': ['*']}

install_requires = \
['jax>=0.1.59,<0.2.0', 'jaxlib>=0.1.40,<0.2.0', 'scikit-learn>=0.23.0,<0.24.0']

setup_kwargs = {
    'name': 'sklearn-jax-kernels',
    'version': '0.0.2',
    'description': 'Composable kernels for scikit-learn implemented in JAX.',
    'long_description': '# sklearn-jax-kernels\n\n[![Build Status](https://travis-ci.com/ExpectationMax/sklearn-jax-kernels.svg?token=3sUUnmMzs9wxN3Qapssj&branch=master)](https://travis-ci.com/ExpectationMax/sklearn-jax-kernels)\n\n**Warning: This project is still in an early stage it could be that the API\nwill change in the future, further functionality is still very limited to the\nuse cases which defined the creation of the project (application to DNA\nsequences present in Biology).**\n\n## Why?\nEver wanted to run a kernel-based model from\n[scikit-learn](https://scikit-learn.org/) on a relatively large dataset?  If so\nyou will have noticed, that this can take extraordinarily long and require huge\namounts of memory, especially if you are using compositions of kernels (such as\nfor example `k1 * k2 + k3`).  This is due to the way Kernels are computed in\nscikit-learn: For each kernel, the complete kernel matrix is computed, and the\ncompositions are then computed from the kernel matrices.  Further,\n`scikit-learn` does not rely on an automatic differentiation framework for the\ncomputation of gradients though kernel operations.\n\n## Introduction\n\n`sklearn-jax-kernels` was designed to circumvent these issues:\n\n - The utilization of [JAX](https://github.com/google/jax) allows accelerating\n   kernel computations through [XLA](https://www.tensorflow.org/xla)\n   optimizations, computation on GPUs and simplifies the computation of\n   gradients though kernels\n - The composition of kernels takes place on a per-element basis, such that\n   unnecessary copies can be optimized away by JAX compilation\n\nThe goal of `sklearn-jax-kernels` is to provide the same flexibility and ease\nof use as known from `scikit-learn` kernels while improving speed and allowing\nthe faster design of new kernels through Automatic Differentiation.\n\nThe kernels in this package follow the [scikit-learn kernel\nAPI](https://scikit-learn.org/stable/modules/gaussian_process.html#gaussian-process-kernel-api).\n\n## Quickstart\n\nA short demonstration of how the kernels can be used, inspired by the\n[ scikit-learn\ndocumentation](https://scikit-learn.org/stable/auto_examples/gaussian_process/plot_gpc_iris.html).\n\n```python\nfrom sklearn import datasets\nimport jax.numpy as jnp\nfrom sklearn_jax_kernels import RBF, GaussianProcessClassifier\n\niris = datasets.load_iris()\nX = jnp.asarray(iris.data)\ny = jnp.array(iris.target, dtype=int)\n\nkernel = 1. + RBF(length_scale=1.0)\ngpc = GaussianProcessClassifier(kernel=kernel).fit(X, y)\n```\n\nHere a further example demonstrating how kernels can be combined:\n\n```python\nfrom sklearn_jax_kernels.base_kernels import RBF, NormalizedKernel\nfrom sklearn_jax_kernels.structured.strings import SpectrumKernel\n\nmy_kernel = RBF(1.) * SpectrumKernel(n_gram_length=3)\nmy_kernel_2 = RBF(1.) + RBF(2.)\nmy_kernel_2 = NormalizedKernel(my_kernel_2)\n```\n\nSome further inspiration can be taken from the tests in the subfolder `tests`.\n\n## Implemented Kernels\n\n - Kernel compositions ($+,-,*,/$, exponentiation)\n - Kernels for real valued data:  \n     - RBF kernel\n - Kernels for same length strings:  \n     - SpectrumKernel\n     - DistanceSpectrumKernel, SpectrumKernel with distance weight between\n       matching substrings\n     - ReverseComplement Spectrum kernel (relevant for applications in Biology\n       when working with DNA sequences)\n\n## TODOs\n\n - Implement more fundamental Kernels\n - Implement jax compatible version of GaussianProcessRegressor\n - Optimize GaussianProcessClassifier for performance\n - Run benchmarks to show benefits in speed\n - Add fake "split" kernel which allows to apply different kernels to different\n   parts of the input\n',
    'author': 'Max Horn',
    'author_email': 'max.horn@bsse.ethz.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ExpectationMax/sklearn-jax-kernels',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Base classes of Kernel implementation compatible with JAX."""
import abc
import numpy
from functools import partial
from sklearn.gaussian_process.kernels import Kernel as sklearn_kernel
from sklearn.gaussian_process.kernels import (
    Hyperparameter,
    StationaryKernelMixin,
    NormalizedKernelMixin
)
from jax import jit, vmap, value_and_grad
import jax.numpy as np
import jax.ops as ops
from jax.experimental import loops

from .config import config_value


class Kernel(sklearn_kernel, metaclass=abc.ABCMeta):
    """Kernel object similar to sklearn implementation but supporting JAX.

    Contains additional methods:
     - pure_kernel_function

    """

    @property
    @abc.abstractmethod
    def pure_kernel_fn(self):
        """Get a pure function which applies the kernel.

        Returned function should have the signature:
        fn(theta, x1, x2).
        """

    @staticmethod
    def _kernel_matrix_without_gradients(kernel_fn, theta, X, Y):
        kernel_fn = partial(kernel_fn, theta)
        if Y is None or (Y is X):
            if config_value('KERNEL_MATRIX_USE_LOOP'):
                n = len(X)
                with loops.Scope() as s:
                    # s.scattered_values = np.empty((n, n))
                    s.index1, s.index2 = np.tril_indices(n, k=0)
                    s.output = np.empty(len(s.index1))
                    for i in s.range(s.index1.shape[0]):
                        i1, i2 = s.index1[i], s.index2[i]
                        s.output = ops.index_update(
                            s.output,
                            i,
                            kernel_fn(X[i1], X[i2])
                        )
                first_update = ops.index_update(
                    np.empty((n, n)), (s.index1, s.index2), s.output)
                second_update = ops.index_update(
                    first_update, (s.index2, s.index1), s.output)
                return second_update
            else:
                n = len(X)
                values_scattered = np.empty((n, n))
                index1, index2 = np.tril_indices(n, k=-1)
                inst1, inst2 = X[index1], X[index2]
                values = vmap(kernel_fn)(inst1, inst2)
                values_scattered = ops.index_update(
                    values_scattered, (index1, index2), values)
                values_scattered = ops.index_update(
                    values_scattered, (index2, index1), values)
                values_scattered = ops.index_update(
                    values_scattered,
                    np.diag_indices(n),
                    vmap(lambda x: kernel_fn(x, x))(X)
                )
                return values_scattered
        else:
            if config_value('KERNEL_MATRIX_USE_LOOP'):
                with loops.Scope() as s:
                    s.output = np.empty((X.shape[0], Y.shape[0]))
                    for i in s.range(X.shape[0]):
                        x = X[i]
                        s.output = ops.index_update(
                            s.output,
                            i,
                            vmap(lambda y: kernel_fn(x, y))(Y)
                        )
                return s.output
            else:
                return vmap(lambda x: vmap(lambda y: kernel_fn(x, y))(Y))(X)

    @staticmethod
    def _kernel_matrix_with_gradients(kernel_fn, theta, X, Y):
        kernel_fn = value_and_grad(kernel_fn)
        kernel_fn = partial(kernel_fn, theta)
        if Y is None or (Y is X):
            if config_value('KERNEL_MATRIX_USE_LOOP'):
                n = len(X)
                with loops.Scope() as s:
                    s.scattered_values = np.empty((n, n))
                    s.scattered_grads = np.empty((n, n, len(theta)))
                    index1, index2 = np.tril_indices(n, k=0)
                    for i in s.range(index1.shape[0]):
                        i1, i2 = index1[i], index2[i]
                        value, grads = kernel_fn(X[i1], X[i2])
                        indexes = (np.stack([i1, i2]), np.stack([i2, i1]))
                        s.scattered_values = ops.index_update(
                            s.scattered_values,
                            indexes,
                            value
                        )
                        s.scattered_grads = ops.index_update(
                            s.scattered_grads, indexes, grads)
                return s.scattered_values, s.scattered_grads
            else:
                n = len(X)
                values_scattered = np.empty((n, n))
                grads_scattered = np.empty((n, n, len(theta)))
                index1, index2 = np.tril_indices(n, k=-1)
                inst1, inst2 = X[index1], X[index2]
                values, grads = vmap(kernel_fn)(inst1, inst2)
                # Scatter computed values into matrix
                values_scattered = ops.index_update(
                    values_scattered, (index1, index2), values)
                values_scattered = ops.index_update(
                    values_scattered, (index2, index1), values)
                grads_scattered = ops.index_update(
                    grads_scattered, (index1, index2), grads)
                grads_scattered = ops.index_update(
                    grads_scattered, (index2, index1), grads)
                diag_values, diag_grads = vmap(
                    lambda x: kernel_fn(x, x))(X)
                diag_indices = np.diag_indices(n)
                values_scattered = ops.index_update(
                    values_scattered, diag_indices, diag_values)
                grads_scattered = ops.index_update(
                    grads_scattered, diag_indices, diag_grads)
                return values_scattered, grads_scattered
        else:
            return vmap(
                lambda x: vmap(lambda y: kernel_fn(x, y))(Y))(X)

    def get_kernel_matrix_fn(self, eval_gradient):
        """Return pure function for computing kernel matrix and gradients.

        We do some internal caching in order to avoid recompiling the resulting
        function.

        Returned function has the signature: `f(theta, X, Y)`
        """
        cache_name = (
            '_cached_kernel_matrix_fn' + '_grad' if eval_gradient else '')
        if not hasattr(self, cache_name):
            pure_kernel_fn = self.pure_kernel_fn

            if eval_gradient:
                kernel_matrix_fn = jit(partial(
                    self._kernel_matrix_with_gradients,
                    pure_kernel_fn
                ))
            else:
                kernel_matrix_fn = jit(partial(
                    self._kernel_matrix_without_gradients,
                    pure_kernel_fn
                ))
            setattr(self, cache_name, kernel_matrix_fn)

        return getattr(self, cache_name)

    def __call__(self, X, Y=None, eval_gradient=False):
        """Build kernel matrix from input data X and Y.

        Evtl. also compute the gradient with respect to the parameters.
        """
        X = np.asarray(X)
        if Y is not None:
            Y = np.asarray(Y)
        return self.get_kernel_matrix_fn(eval_gradient)(self.theta, X, Y)

    def diag(self, X):
        """Get diagonal of kernel matrix."""
        return vmap(lambda x: self.pure_kernel_fn(self.theta, x, x))(X)

    def __add__(self, b):
        """Add kernel to constant or other kernel."""
        if not isinstance(b, Kernel):
            return Sum(self, ConstantKernel(b))
        return Sum(self, b)

    def __radd__(self, b):
        """Add kernel to constant or other kernel."""
        if not isinstance(b, Kernel):
            return Sum(ConstantKernel(b), self)
        return Sum(b, self)

    def __mul__(self, b):
        """Mulitply kernel with constant or other kernel."""
        if not isinstance(b, Kernel):
            return Product(self, ConstantKernel(b))
        return Product(self, b)

    def __rmul__(self, b):
        """Mulitply kernel with constant or other kernel."""
        if not isinstance(b, Kernel):
            return Product(ConstantKernel(b), self)
        return Product(b, self)

    def __pow__(self, b):
        """Exponentiate kernel."""
        return Exponentiation(self, b)


class NormalizedKernel(NormalizedKernelMixin, Kernel):
    """Kernel wrapper which computes a normalized version of the kernel."""

    def __init__(self, kernel):
        self.kernel = kernel

    @property
    def pure_kernel_fn(self):
        """Not really needed in this particular case."""
        kernel_fn = self.kernel.pure_kernel_fn

        def wrapped(theta, x, y):
            K_xy = kernel_fn(theta, x, y)
            K_xx = kernel_fn(theta, x, x)
            K_yy = kernel_fn(theta, y, y)
            return K_xy / np.sqrt(K_xx * K_yy)
        return wrapped

    def get_kernel_matrix_fn(self, eval_gradient):
        """Return pure function for computing kernel matrix and gradients.

        We do some internal caching in order to avoid recompiling the resulting
        function. Further, we compute the output of the normalized kernel
        matrix at this stage in order to avoid recomputing the self similarity
        on each kernel evaluation.

        Returned function has the signature: `f(theta, X, Y)`
        """
        if config_value('NORMALIZED_KERNEL_PUSH_DOWN'):
            # In this case compute the normalization for each instance
            # inside the kernel fn. This recomputes the self similarities many
            # times, but does not require keeping multiple tensors of the size
            # of the kernel matrix in memory for computing normalization. This
            # is particularly the case when computing gradients with respect to
            # kernel parameters.
            return super().get_kernel_matrix_fn(eval_gradient)

        cache_name = '_kernel_matrix_fn' + '_grad' if eval_gradient else ''
        if not hasattr(self, cache_name):
            pure_kernel_fn = self.kernel.pure_kernel_fn
            if eval_gradient:
                kernel_matrix_with_grad = \
                    self._kernel_matrix_with_gradients

                def wrapped(theta, X, Y):
                    """Compute normalized kernel matrix and do chain rule."""
                    kmatrix, grads = kernel_matrix_with_grad(
                        pure_kernel_fn, theta, X, Y)
                    if Y is None:
                        diag = np.diag(kmatrix)
                        grad_diag_indices = np.diag_indices(kmatrix.shape[0])
                        diag_grad = grads[grad_diag_indices]
                        normalizer = np.sqrt(diag[:, None] * diag[None, :])
                        # Add dimensions for broadcasting
                        K_xx = diag[:, None, None]
                        K_yy = diag[None, :, None]
                        K_xx_grad = diag_grad[:, None, :]
                        K_yy_grad = diag_grad[None, :, :]

                        # Do the chain rule
                        grads = (
                            (
                                2 * K_xx * K_yy * grads -
                                kmatrix * (K_xx_grad * K_yy + K_xx * K_yy_grad)
                            ) / (2 * (K_xx * K_yy) ** (3/2))
                        )
                        return kmatrix / normalizer, grads
                    else:
                        # If y is not defined we need to compute the self
                        # similarity of each instance
                        kernel_fn_with_grad = partial(
                            value_and_grad(pure_kernel_fn), theta)
                        K_xx, K_xx_grad = vmap(
                            lambda x: kernel_fn_with_grad(x, x))(X)
                        K_yy, K_yy_grad = vmap(
                            lambda y: kernel_fn_with_grad(y, y))(Y)
                        # Add dimensions for broadcasting
                        K_xx = K_xx[:, None, None]
                        K_yy = K_yy[None, :, None]
                        K_xx_grad = K_xx_grad[:, None, :]
                        K_yy_grad = K_yy_grad[None, :, :]

                        normalizer = np.sqrt(K_xx[:, None] * K_yy[None, :])
                        # d/dw(k(x, y, w)/sqrt(k(x, x, w) k(y, y, w))) = (2
                        # k(x, x, w) k(y, y, w) k^(0, 0, 1)(x, y, w) - k(x, y,
                        # w) (k^(0, 0, 1)(x, x, w) k(y, y, w) + k(x, x, w)
                        # k^(0, 0, 1)(y, y, w)))/(2 (k(x, x, w) k(y, y,
                        # w))^(3/2))
                        grads = (
                            (
                                2 * K_xx * K_yy * grads -
                                kmatrix * (K_xx_grad * K_yy + K_xx * K_yy_grad)
                            ) /
                            (2 * (K_xx * K_yy) ** (3/2))
                        )

                        return kmatrix / normalizer, grads

                kernel_matrix_fn = jit(wrapped)
            else:
                kernel_matrix = self._kernel_matrix_without_gradients

                def wrapped(theta, X, Y):
                    """Compute normalized kernel matrix."""
                    kmatrix = kernel_matrix(pure_kernel_fn, theta, X, Y)
                    if Y is None:
                        diag = np.diag(kmatrix)
                        normalizer = np.sqrt(diag[:, None] * diag[None, :])
                        return kmatrix / normalizer
                    else:
                        # If y is not defined we need to compute the self
                        # similarity of each instance
                        K_xx = vmap(lambda x: pure_kernel_fn(theta, x, x))(X)
                        K_yy = vmap(lambda y: pure_kernel_fn(theta, y, y))(Y)
                        normalizer = np.sqrt(K_xx[:, None] * K_yy[None, :])
                        return kmatrix / normalizer

                kernel_matrix_fn = jit(wrapped)
            setattr(self, cache_name, kernel_matrix_fn)
        return getattr(self, cache_name)

    def is_stationary(self):
        """Whether kernel is stationary."""
        return self.kernel.is_stationary()

    def get_params(self, deep=True):
        """Get parameters of this kernel.

        Parameters:
            deep : boolean, optional
                If True, will return the parameters for this estimator and
                contained subobjects that are estimators.

        Returns:
            params : mapping of string to any
                Parameter names mapped to their values.

        """
        params = dict(kernel=self.kernel)
        if deep:
            deep_items = self.kernel.get_params().items()
            params.update(('kernel__' + k, val) for k, val in deep_items)
        return params

    @property
    def hyperparameters(self):
        """Return a list of all hyperparameter."""
        r = []
        for hyperparameter in self.kernel.hyperparameters:
            r.append(Hyperparameter("kernel__" + hyperparameter.name,
                                    hyperparameter.value_type,
                                    hyperparameter.bounds,
                                    hyperparameter.n_elements))
        return r

    @property
    def theta(self):
        """Return the (flattened, log-transformed) non-fixed hyperparameters.

        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.

        Returns:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        return self.kernel.theta

    @theta.setter
    def theta(self, theta):
        """Set the (flattened, log-transformed) non-fixed hyperparameters.

        Parameters:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        self.kernel.theta = theta

    @property
    def bounds(self):
        """Return the log-transformed bounds on the theta.

        Returns:
            bounds : array, shape (n_dims, 2)
                The log-transformed bounds on the kernel's hyperparameters
                theta

        """
        return self.kernel.bounds

    def __eq__(self, b):
        """Whether two instances are considered equal."""
        if type(self) != type(b):
            return False
        return self.kernel == b.kernel


class KernelOperator(Kernel):
    """Base class for all kernel operators."""

    def __init__(self, k1, k2):
        self.k1 = k1
        self.k2 = k2

    def get_params(self, deep=True):
        """Get parameters of this kernel.

        Parameters:
            deep : boolean, optional
                If True, will return the parameters for this estimator and
                contained subobjects that are estimators.

        Returns:
            params : mapping of string to any
                Parameter names mapped to their values.

        """
        params = dict(k1=self.k1, k2=self.k2)
        if deep:
            deep_items = self.k1.get_params().items()
            params.update(('k1__' + k, val) for k, val in deep_items)
            deep_items = self.k2.get_params().items()
            params.update(('k2__' + k, val) for k, val in deep_items)

        return params

    @property
    def hyperparameters(self):
        """Return a list of all hyperparameter."""
        r = [Hyperparameter("k1__" + hyperparameter.name,
                            hyperparameter.value_type,
                            hyperparameter.bounds, hyperparameter.n_elements)
             for hyperparameter in self.k1.hyperparameters]

        for hyperparameter in self.k2.hyperparameters:
            r.append(Hyperparameter("k2__" + hyperparameter.name,
                                    hyperparameter.value_type,
                                    hyperparameter.bounds,
                                    hyperparameter.n_elements))
        return r

    @property
    def theta(self):
        """Return the (flattened, log-transformed) non-fixed hyperparameters.

        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.

        Returns:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        return np.append(self.k1.theta, self.k2.theta)

    @theta.setter
    def theta(self, theta):
        """Set the (flattened, log-transformed) non-fixed hyperparameters.

        Parameters:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        k1_dims = self.k1.n_dims
        self.k1.theta = theta[:k1_dims]
        self.k2.theta = theta[k1_dims:]

    @property
    def bounds(self):
        """Return the log-transformed bounds on the theta.

        Returns:
            bounds : array, shape (n_dims, 2)
                The log-transformed bounds on the kernel's hyperparameters
                theta

        """
        if self.k1.bounds.size == 0:
            return self.k2.bounds
        if self.k2.bounds.size == 0:
            return self.k1.bounds
        return np.vstack((self.k1.bounds, self.k2.bounds))

    def __eq__(self, b):
        """Check for equality between kernels."""
        if type(self) != type(b):
            return False
        return (self.k1 == b.k1 and self.k2 == b.k2) \
            or (self.k1 == b.k2 and self.k2 == b.k1)

    def is_stationary(self):
        """Return whether the kernel is stationary."""
        return self.k1.is_stationary() and self.k2.is_stationary()

    @property
    def requires_vector_input(self):
        """Return whether the kernel is stationary. """
        return (self.k1.requires_vector_input or
                self.k2.requires_vector_input)


class Sum(KernelOperator):
    """Sum of two kernels."""

    @property
    def pure_kernel_fn(self):
        """Kernel function of the two added kernels."""
        k1_fn = self.k1.pure_kernel_fn
        k2_fn = self.k2.pure_kernel_fn
        k1_dims = self.k1.n_dims

        def kernel_fn(theta, x, y):
            return k1_fn(theta[:k1_dims], x, y) + k2_fn(theta[k1_dims:], x, y)

        return kernel_fn


class Product(KernelOperator):
    """Product of two kernels."""

    @property
    def pure_kernel_fn(self):
        """Kernel function of the two added kernels."""
        k1_fn = self.k1.pure_kernel_fn
        k2_fn = self.k2.pure_kernel_fn
        k1_dims = self.k1.n_dims

        def kernel_fn(theta, x, y):
            return k1_fn(theta[:k1_dims], x, y) * k2_fn(theta[k1_dims:], x, y)

        return kernel_fn

    def __repr__(self):
        """Return representation of kernel."""
        return "{0} * {1}".format(self.k1, self.k2)


class ConstantKernel(StationaryKernelMixin, Kernel):
    """Kernel which always returns a constant."""

    def __init__(self, constant_value=1.0, constant_value_bounds=(1e-5, 1e5)):
        """Init kernel with constant_value."""
        self.constant_value = constant_value
        self.constant_value_bounds = constant_value_bounds

    @property
    def hyperparameter_constant_value(self):
        return Hyperparameter(
            "constant_value", "numeric", self.constant_value_bounds)

    @property
    def pure_kernel_fn(self):
        """Return the kernel fn."""
        if self.hyperparameter_constant_value.fixed:
            value = self.constant_value

            def kernel_fn(theta, x, y):
                return value
        else:
            def kernel_fn(theta, x, y):
                return np.exp(theta[0])  # Theta is in log domain and array

        return kernel_fn


class Exponentiation(Kernel):
    """Exponentiation of a kernel."""

    def __init__(self, kernel: Kernel, exponent):
        """Init kernel exponentiation of kernel with exponent."""
        self.kernel = kernel
        self.exponent = exponent

    def is_stationary(self):
        """Whether kernel is stationary."""
        return self.kernel.is_stationary()

    def get_params(self, deep=True):
        """Get parameters of this kernel.

        Parameters:
            deep : boolean, optional
                If True, will return the parameters for this estimator and
                contained subobjects that are estimators.

        Returns:
            params : mapping of string to any
                Parameter names mapped to their values.

        """
        params = dict(kernel=self.kernel, exponent=self.exponent)
        if deep:
            deep_items = self.kernel.get_params().items()
            params.update(('kernel__' + k, val) for k, val in deep_items)
        return params

    @property
    def hyperparameters(self):
        """Return a list of all hyperparameter."""
        r = []
        for hyperparameter in self.kernel.hyperparameters:
            r.append(Hyperparameter("kernel__" + hyperparameter.name,
                                    hyperparameter.value_type,
                                    hyperparameter.bounds,
                                    hyperparameter.n_elements))
        return r

    @property
    def theta(self):
        """Return the (flattened, log-transformed) non-fixed hyperparameters.

        Note that theta are typically the log-transformed values of the
        kernel's hyperparameters as this representation of the search space
        is more amenable for hyperparameter search, as hyperparameters like
        length-scales naturally live on a log-scale.

        Returns:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        return self.kernel.theta

    @theta.setter
    def theta(self, theta):
        """Set the (flattened, log-transformed) non-fixed hyperparameters.

        Parameters:
            theta : array, shape (n_dims,)
                The non-fixed, log-transformed hyperparameters of the kernel

        """
        self.kernel.theta = theta

    @property
    def bounds(self):
        """Return the log-transformed bounds on the theta.

        Returns:
            bounds : array, shape (n_dims, 2)
                The log-transformed bounds on the kernel's hyperparameters
                theta

        """
        return self.kernel.bounds

    def __eq__(self, b):
        """Whether two instances are considered equal."""
        if type(self) != type(b):
            return False
        return (self.kernel == b.kernel and self.exponent == b.exponent)

    @property
    def pure_kernel_fn(self):
        """Pure kernel fn of exponentiated kernel."""
        exponent = self.exponent
        kernel_fn = self.kernel.pure_kernel_fn

        def exp_kernel_fn(theta, x, y):
            return np.power(kernel_fn(theta, x, y), exponent)
        return exp_kernel_fn


class RBF(StationaryKernelMixin, NormalizedKernelMixin, Kernel):
    """RBF Kernel."""

    def __init__(self, length_scale=1.0, length_scale_bounds=(1e-5, 1e5)):
        """Initialize RBF kernel with length_scale and bounds."""
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds

    @property
    def anisotropic(self):
        return np.iterable(self.length_scale) and len(self.length_scale) > 1

    @property
    def hyperparameter_length_scale(self):
        if self.anisotropic:
            return Hyperparameter("length_scale", "numeric",
                                  self.length_scale_bounds,
                                  len(self.length_scale))
        return Hyperparameter(
            "length_scale", "numeric", self.length_scale_bounds)

    @property
    def pure_kernel_fn(self):
        """Pure kernel fn of RBF kernel."""
        if self.hyperparameter_length_scale.fixed:
            length_scale = self.length_scale
            if np.iterable(length_scale):
                # handle case when length scale is fixed and provided as list
                length_scale = np.asarray(length_scale)

            def kernel_fn(theta, x, y):
                # as we get a log-transformed theta as input, we need to transform
                # it back.
                diff = (x - y) / length_scale
                d = np.sum(diff ** 2, axis=-1)
                return np.exp(-0.5 * d)
        else:
            def kernel_fn(theta, x, y):
                # as we get a log-transformed theta as input, we need to transform
                # it back.
                diff = (x - y) / np.exp(theta)
                d = np.sum(diff ** 2, axis=-1)
                return np.exp(-0.5 * d)
        return kernel_fn

    def __repr__(self):
        if self.anisotropic:
            return "{0}(length_scale=[{1}])".format(
                self.__class__.__name__, ", ".join(map("{0:.3g}".format,
                                                   self.length_scale)))
        else:  # isotropic
            return "{0}(length_scale={1:.3g})".format(
                self.__class__.__name__, np.ravel(self.length_scale)[0])
